#include <stdint.h>
#include <stdio.h>
#include <propeller2.h>
#include "tjpgd.c"

typedef uint16_t pixel_t;
typedef struct __using("p2videodrv.spin2") vdrv_t;

#define JDEC_WORKBUF_SIZE (10*1024)

vdrv_t vdrv;
uint8_t display[vdrv.DISPLAY_SIZE];
uint8_t region[vdrv.REGION_SIZE];
pixel_t lineBuffers[640*2];
pixel_t screenBuffer[320*240];
uint8_t workBuf[JDEC_WORKBUF_SIZE];
uint8_t vcog;

size_t in_func (    /* Returns number of bytes read (zero on error) */
    JDEC* jd,       /* Decompression object */
    uint8_t* buff,  /* Pointer to the read buffer (null to remove data) */
    size_t nbyte    /* Number of bytes to read/remove */
) {
    if (buff) memcpy(buff,jd->device,nbyte);
    jd->device = ((uint8_t*)jd->device)+nbyte;
    return nbyte;
}

int out_func (
  JDEC* jdec,    /* Pointer to the decompression object */
  void* bitmap,  /* Bitmap to be output */
  JRECT* rect    /* Rectangle to output */
) {
    pixel_t *bmp = bitmap;
    pixel_t *out = screenBuffer + (rect->left+rect->top*320);
    unsigned width = rect->right - rect->left + 1;
    unsigned height = rect->bottom - rect->top + 1;

    #if 1
    unsigned tmp;
    __asm volatile {
            shr width,#1 wcz
            rdfast #0,bmp
    .linelp
            mov ptrb,out
            add out,##320*sizeof(pixel_t)
      if_nz rep @.pixlp,width
      if_nz rflong tmp
      if_nz wrlong tmp,ptrb++
    .pixlp
      if_c  rfword tmp
      if_c  wrword tmp,ptrb++
            djnz height,#.linelp
    }
    #else
    do {
        pixel_t *line = out;
        uint16_t x=width;
        do {
            *line = *bmp;
            *line++;
            *bmp++;
        } while (--x);
        out += 320;
    } while (--height);
    #endif
    return 1;
}

uint32_t do_decomp(uint8_t *file) {
    uint32_t time = _cnt();
    JDEC jdec = {};
    jd_prepare(&jdec,&in_func,workBuf,JDEC_WORKBUF_SIZE,file);
    jd_decomp(&jdec,&out_func,0);
    return _cnt() - time;
}

void main() {

    vcog = vdrv.initDisplay( -1,           // the cogid to use (-1 = auto-allocate)
                            display,       // the display structure address in HUB RAM
                            vdrv.VGA,       // video output type (VGA/DVI etc)
                            32,  // base pin number (hsync pin) of VGA pin group
                            32+4,    // VSYNC pin (not used for DVI)
                            0,              // display flags
                            lineBuffers, // address of the consecutive two scan line buffers in HUB RAM
                            640*sizeof(pixel_t),     // size of a single scan line buffer in bytes
                            vdrv.getTiming(vdrv.RES_640x480), // obtain stock timing to use, (or create custom timing instead)
                            0, // optional external memory mailbox address in HUB RAM (0=none)
                            vdrv.initRegion( // setup a single gfx region as the display list
                                        region,        // region structure address in HUB RAM
                                        vdrv.RGB16,       //
                                        0,              // size of region in scan lines (0=to end of screen)
                                        vdrv.DOUBLE_WIDE+vdrv.DOUBLE_HIGH, // region specific flags (if enabled text flashes if BG colour > 7)
                                        0,       // address of default palette to be used by region
                                        0,              // address of default font to be used by this region
                                        8,              // number of scan lines in font
                                        screenBuffer,    // address of screen buffer in HUB RAM
                                        0)              // link to next region, NULL = last region
                            );
    for (;;) {
        unsigned total = 0;
        #define DOTEST(file) {unsigned onetime = do_decomp(&file);total+=onetime;printf(#file " decompd in %u cycles!\n",onetime);}
        DOTEST(reimu_jpg);
        DOTEST(pigge_jpg);
        DOTEST(edge_jpg);
        DOTEST(youmu_jpg);
        printf("TOTAL test suite: %u cycles!!!!\n",total);
    }
}

__asm {
reimu_jpg file "reimu240p.jpg"
pigge_jpg file "pigge240p.jpg"
edge_jpg file "edge240p.jpg"
youmu_jpg file "youmu240p.jpg"
}
